from typing import Generic, TypeVar

from rainbowadn.core import *
from .states import *

__all__ = ('BlockChainProtocol',)

HeaderType = TypeVar('HeaderType')
StateType = TypeVar('StateType')
ActualStateType = TypeVar('ActualStateType')


class BlockChainProtocol(
    Generic[HeaderType, StateType, ActualStateType],
):
    def __init__(
            self,
            protocol: ActiveStateProtocol[HeaderType, StateType],
            header_factory: RainbowFactory[StateType],
            state_factory: RainbowFactory[StateType]
    ):
        assert isinstance(protocol, ActiveStateProtocol)
        assert isinstance(header_factory, RainbowFactory)
        assert isinstance(state_factory, RainbowFactory)
        self.state_protocol = protocol
        self.header_factory = header_factory
        self.state_factory = state_factory

    def actual_state_factory(self) -> RainbowFactory[ActualStateType]:
        raise NotImplementedError

    async def actual_state(
            self,
            state: StateType
    ) -> HashPoint[ActualStateType]:
        raise NotImplementedError

    def loose(self) -> 'BlockChainProtocol[HeaderType, StateType, ActualStateType]':
        return self
