from typing import Generic, TypeVar

from rainbowadn.collection.collectioninterface import CollectionInterface
from rainbowadn.core import *
from rainbowadn.nullability import *

__all__ = ('BlockCollectionInterface',)

BlockType = TypeVar('BlockType')
HeaderType = TypeVar('HeaderType')
ActualStateType = TypeVar('ActualStateType')


class BlockCollectionInterface(
    CollectionInterface[
        BlockType
    ],
    Generic[BlockType, HeaderType, ActualStateType]
):
    def __init__(
            self,
            reference: NullableReference[BlockType]
    ):
        assert isinstance(reference, NullableReference)
        super().__init__(reference)

    async def _verify_link(
            self,
            block: BlockType,
            previous: NullableReference[BlockType]
    ) -> bool:
        raise NotImplementedError

    async def verify_link(
            self,
            previous: NullableReference[BlockType]
    ) -> bool:
        assert isinstance(previous, NullableReference)
        if self.reference.null():
            return True
        else:
            return await self._verify_link(
                await self.reference.resolve(),
                previous
            )

    def _actual_state_factory(self) -> RainbowFactory[ActualStateType]:
        raise NotImplementedError

    async def _actual_state(self, block: BlockType) -> HashPoint[ActualStateType]:
        raise NotImplementedError

    async def actual_state(self) -> NullableReference[ActualStateType]:
        if self.reference.null():
            return NullableReference(Null(), self._actual_state_factory())
        else:
            return NullableReference.of(await self._actual_state(await self.reference.resolve()))

    async def add(
            self,
            header: HashPoint[HeaderType]
    ) -> 'BlockCollectionInterface[BlockType, HeaderType, ActualStateType]':
        raise NotImplementedError
