from typing import TypeVar

from rainbowadn.chain.stages import *
from rainbowadn.chain.states import *
from rainbowadn.core import *
from rainbowadn.nullability import *
from .activestageprotocol import ActiveStageProtocol

__all__ = ('ActiveStageStateProtocol',)

HeaderType = TypeVar('HeaderType')
BaseStateType = TypeVar('BaseStateType')
StageType = TypeVar('StageType')


class ActiveStageStateProtocol(
    ActiveStateProtocol[
        HeaderType,
        StateStage[
            HeaderType,
            BaseStateType,
            StageType
        ]
    ]
):
    def __init__(
            self,
            protocol: ActiveStageProtocol,
            stage_factory: RainbowFactory[StageType],
            base_state_factory: RainbowFactory[BaseStateType],
    ):
        assert isinstance(protocol, ActiveStageProtocol)
        assert isinstance(stage_factory, RainbowFactory)
        assert isinstance(base_state_factory, RainbowFactory)
        self.protocol = protocol
        self.stage_factory = stage_factory
        self.base_state_factory = base_state_factory

    async def verify(
            self,
            previous: NullableReference[
                StateStage[
                    HeaderType,
                    BaseStateType,
                    StageType
                ]
            ],
            header: HashPoint[HeaderType],
            state: HashPoint[
                StateStage[
                    HeaderType,
                    BaseStateType,
                    StageType
                ]
            ]
    ) -> bool:
        assert isinstance(previous, NullableReference)
        assert isinstance(header, HashPoint)
        assert isinstance(state, HashPoint)
        return await self.stage_state_protocol().verify(
            previous,
            header,
            state
        )

    async def derive(
            self,
            previous: NullableReference[
                StateStage[
                    HeaderType,
                    BaseStateType,
                    StageType
                ]
            ],
            header: HashPoint[HeaderType],
    ) -> HashPoint[
        StateStage[
            HeaderType,
            BaseStateType,
            StageType
        ]
    ]:
        assert isinstance(previous, NullableReference)
        assert isinstance(header, HashPoint)
        return await self.protocol.derive_full(
            previous,
            header,
            self.base_state_factory,
            self.stage_factory
        )

    @classmethod
    def stage_state_protocol(cls) -> StateProtocol[
        HeaderType,
        StateStage[HeaderType, BaseStateType, StageType]
    ]:
        return StageStateProtocol()
