from typing import Generic, TypeVar

from rainbowadn.chain.derivation import *
from rainbowadn.core import *
from rainbowadn.nullability import *
from .reduced import Reduced
from .reducible import Reducible
from .reductionprotocol import ReductionProtocol
from .reductionresult import ReductionResult

__all__ = ('ReductionStageProtocol',)

ReductorType = TypeVar('ReductorType')
AccumulatorType = TypeVar('AccumulatorType')


class ReductionStageProtocol(
    ActiveStageProtocol[ReductorType, AccumulatorType, Reducible[ReductorType, AccumulatorType]],
    Generic[ReductorType, AccumulatorType]
):
    def __init__(self, protocol: ReductionProtocol[ReductorType, AccumulatorType]):
        assert isinstance(protocol, ReductionProtocol)
        self.protocol = protocol
        super().__init__()

    async def _derive_accumulator(
            self,
            previous: NullableReference[AccumulatorType],
    ) -> HashPoint[AccumulatorType]:
        assert isinstance(previous, NullableReference)
        if previous.null():
            return self.protocol.initial(previous.factory)
        else:
            return await self.protocol.header_filter(previous.hashpoint())

    async def derive_header(
            self,
            previous: NullableReference[AccumulatorType],
            header: HashPoint[ReductorType]
    ) -> HashPoint[Reducible[ReductorType, AccumulatorType]]:
        assert isinstance(previous, NullableReference)
        assert isinstance(header, HashPoint)
        return HashPoint.of(
            Reducible(
                header,
                await self._derive_accumulator(previous)
            )
        )

    @classmethod
    def _derived_from_reduced(
            cls,
            reduced: ReductionResult[ReductorType, AccumulatorType]
    ) -> Derived[ReductorType, Reducible[ReductorType, AccumulatorType]]:
        assert isinstance(reduced, ReductionResult)
        if isinstance(reduced, Reducible):
            return DerivedStage(HashPoint.of(reduced))
        elif isinstance(reduced, Reduced):
            return DerivedState(reduced.reduced)
        else:
            raise TypeError

    async def derive_stage_or_state(
            self,
            stage: HashPoint[Reducible[ReductorType, AccumulatorType]]
    ) -> Derived[ReductorType, Reducible[ReductorType, AccumulatorType]]:
        assert isinstance(stage, HashPoint)
        return self._derived_from_reduced(await self.protocol.reduce(await stage.resolve()))
