from typing import Generic, TypeVar

from rainbowadn.core import *
from .block import Block
from .blockchain import BlockChainFactory
from .chaincollectionfactory import ChainCollectionFactory
from .reduction import *
from .stages import *

__all__ = ('ReductionChainMetaFactory',)

ReductorType = TypeVar('ReductorType')
AccumulatorType = TypeVar('AccumulatorType')


class ReductionChainMetaFactory(
    AbstractReductionChainMetaFactory[
        Block[
            ReductorType,
            StateStage[ReductorType, AccumulatorType, Reducible[ReductorType, AccumulatorType]]
        ],
        ReductorType,
        AccumulatorType
    ],
    Generic[ReductorType, AccumulatorType]
):
    def factory(
            self,
            reductor_factory: RainbowFactory[ReductorType],
            accumulator_factory: RainbowFactory[AccumulatorType],
            protocol: ReductionProtocol[ReductorType, AccumulatorType],
    ) -> ChainCollectionFactory[
        Block[
            ReductorType,
            StateStage[ReductorType, AccumulatorType, Reducible[ReductorType, AccumulatorType]]
        ],
        ReductorType,
        AccumulatorType
    ]:
        assert isinstance(reductor_factory, RainbowFactory)
        assert isinstance(accumulator_factory, RainbowFactory)
        assert isinstance(protocol, ReductionProtocol)
        return BlockChainFactory(
            ReductionChainProtocol(
                protocol,
                reductor_factory,
                accumulator_factory,
            )
        )

    def loose(self) -> AbstractReductionChainMetaFactory[
        Block[
            ReductorType,
            StateStage[ReductorType, AccumulatorType, Reducible[ReductorType, AccumulatorType]]
        ],
        ReductorType,
        AccumulatorType
    ]:
        return self
