from typing import Generic, TypeVar

from rainbowadn.core import *
from rainbowadn.nullability import *

__all__ = ('StageProtocol',)

HeaderType = TypeVar('HeaderType')
BaseStateType = TypeVar('BaseStateType')
StageType = TypeVar('StageType')


class StageProtocol(Generic[HeaderType, BaseStateType, StageType]):
    async def verify_header(
            self,
            previous: NullableReference[BaseStateType],
            header: HashPoint[HeaderType],
            stage: HashPoint[StageType]
    ) -> bool:
        raise NotImplementedError

    async def verify_stage(
            self,
            previous: HashPoint[StageType],
            stage: HashPoint[StageType]
    ) -> bool:
        raise NotImplementedError

    async def verify_state(
            self,
            stage: HashPoint[StageType],
            state: HashPoint[BaseStateType]
    ) -> bool:
        raise NotImplementedError
