from typing import Generic, TypeVar

from rainbowadn.collection.keyed import Keyed
from rainbowadn.core import *
from .comparator import *

__all__ = ('KeyedComparator',)

ComparatorKeyType = TypeVar('ComparatorKeyType')


class KeyedComparator(
    Comparator[Keyed[ComparatorKeyType]], Generic[ComparatorKeyType]
):
    def __init__(self, comparator: Comparator[ComparatorKeyType]):
        assert isinstance(comparator, Comparator)
        self.comparator = comparator
        super().__init__()

    async def compare(
            self,
            original: HashPoint[Keyed[ComparatorKeyType]],
            key: HashPoint[Keyed[ComparatorKeyType]]
    ) -> Comparison:
        assert isinstance(original, HashPoint)
        assert isinstance(key, HashPoint)
        original_resolved: Keyed[ComparatorKeyType]
        key_resolved: Keyed[ComparatorKeyType]
        original_resolved, key_resolved = await gather(
            original.resolve(),
            key.resolve(),
        )
        assert isinstance(original_resolved, Keyed)
        assert isinstance(key_resolved, Keyed)
        return await self.comparator.compare(
            original_resolved.key,
            key_resolved.key,
        )
