from typing import Generic, TypeVar

from rainbowadn.collection.comparison import *
from rainbowadn.collection.trees.binary.core import *
from .binaryaction import BinaryAction
from .compareaction import CompareAction

__all__ = ('AddAction', 'RemoveAction', 'ContainsAction',)

TreeType = TypeVar('TreeType')
ActiveKeyType = TypeVar('ActiveKeyType')
MetaDataType = TypeVar('MetaDataType')


class AddAction(
    CompareAction[
        ActiveKeyType,
        MetaDataType,
        TreeType,
        TreeType
    ],
    Generic[ActiveKeyType, MetaDataType, TreeType]
):
    async def on_equal(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, MetaDataType, TreeType],
            equal: Equal
    ) -> TreeType:
        assert isinstance(case, ProtocolizedBinarySplit)
        assert isinstance(equal, Equal)
        if isinstance(equal, Replace):
            return await case.protocol.tree(case.split.treel, case.split.treer, self.key)
        else:
            raise TypeError

    async def on_left(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> TreeType:
        assert isinstance(case, ProtocolizedBinarySplit)
        return await case.protocol.tree(
            await self.on(case.protocolizedl()),
            case.split.treer,
            case.split.key
        )

    async def on_right(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> TreeType:
        assert isinstance(case, ProtocolizedBinarySplit)
        return await case.protocol.tree(
            case.split.treel,
            await self.on(case.protocolizedr()),
            case.split.key
        )

    async def on_null(
            self,
            protocolized: BinaryProtocolized[ActiveKeyType, MetaDataType, TreeType],
    ) -> TreeType:
        assert isinstance(protocolized, BinaryProtocolized)
        return await protocolized.creation.tree(protocolized.tree, protocolized.tree, self.key)


class MergeAction(
    BinaryAction[
        ActiveKeyType,
        MetaDataType,
        TreeType,
        TreeType
    ],
    Generic[ActiveKeyType, MetaDataType, TreeType]
):
    def __init__(self, treer: TreeType):
        self.treer = treer

    async def on_null(
            self,
            protocolized: BinaryProtocolized[ActiveKeyType, MetaDataType, TreeType],
    ) -> TreeType:
        assert isinstance(protocolized, BinaryProtocolized)
        return self.treer

    async def on_split(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> TreeType:
        assert isinstance(case, ProtocolizedBinarySplit)
        return await case.protocol.tree(
            case.split.treel,
            await MergeAction(self.treer).on(case.protocolizedr()),
            case.split.key
        )


class RemoveAction(
    CompareAction[
        ActiveKeyType,
        MetaDataType,
        TreeType,
        TreeType
    ],
    Generic[ActiveKeyType, MetaDataType, TreeType]
):
    async def on_equal(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, MetaDataType, TreeType],
            equal: Equal
    ) -> TreeType:
        assert isinstance(case, ProtocolizedBinarySplit)
        assert isinstance(equal, Equal)
        return await MergeAction(case.split.treer).on(case.protocolizedl())

    async def on_left(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> TreeType:
        assert isinstance(case, ProtocolizedBinarySplit)
        return await case.protocol.tree(
            await self.on(case.protocolizedl()),
            case.split.treer,
            case.split.key
        )

    async def on_right(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> TreeType:
        assert isinstance(case, ProtocolizedBinarySplit)
        return await case.protocol.tree(
            case.split.treel,
            await self.on(case.protocolizedr()),
            case.split.key
        )

    async def on_null(
            self,
            protocolized: BinaryProtocolized[ActiveKeyType, MetaDataType, TreeType],
    ) -> TreeType:
        assert isinstance(protocolized, BinaryProtocolized)
        return protocolized.tree


class ContainsAction(
    CompareAction[
        ActiveKeyType,
        MetaDataType,
        TreeType,
        bool
    ],
    Generic[ActiveKeyType, MetaDataType, TreeType]
):
    async def on_equal(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, MetaDataType, TreeType],
            equal: Equal
    ) -> bool:
        assert isinstance(case, ProtocolizedBinarySplit)
        assert isinstance(equal, Equal)
        return True

    async def on_left(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> bool:
        assert isinstance(case, ProtocolizedBinarySplit)
        return await self.on(case.protocolizedl())

    async def on_right(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, MetaDataType, TreeType]
    ) -> bool:
        assert isinstance(case, ProtocolizedBinarySplit)
        return await self.on(case.protocolizedr())

    async def on_null(
            self,
            protocolized: BinaryProtocolized[ActiveKeyType, MetaDataType, TreeType],
    ) -> bool:
        assert isinstance(protocolized, BinaryProtocolized)
        return False
