import abc
from typing import Generic, TypeVar

from rainbowadn.core import *
from .binarybalancing import BinaryBalancing
from .binarycreation import BinaryCreation
from .binaryprotocolized import BinaryProtocolized

__all__ = ('BalancedCreation',)

TreeType = TypeVar('TreeType')
ActiveKeyType = TypeVar('ActiveKeyType')
MetaDataType = TypeVar('MetaDataType')


class BalancedCreation(
    BinaryCreation[ActiveKeyType, MetaDataType, TreeType],
    Generic[ActiveKeyType, MetaDataType, TreeType],
    abc.ABC
):
    def __init__(
            self,
            protocol: BinaryBalancing[ActiveKeyType, MetaDataType, TreeType]
    ):
        assert isinstance(protocol, BinaryBalancing)
        self.protocol = protocol
        super().__init__(protocol.comparator)

    async def _tree(self, treel: TreeType, treer: TreeType, key: HashPoint[ActiveKeyType]) -> TreeType:
        raise NotImplementedError

    async def tree(self, treel: TreeType, treer: TreeType, key: HashPoint[ActiveKeyType]) -> TreeType:
        assert isinstance(key, HashPoint)
        balancedl, balancedr = await gather(
            self.protocol.balance(BinaryProtocolized(self, treel)),
            self.protocol.balance(BinaryProtocolized(self, treer)),
        )
        return await self.protocol.balance(
            BinaryProtocolized(
                self,
                await self._tree(
                    balancedl,
                    balancedr,
                    key
                )
            )
        )
