import abc
from typing import TypeVar

from .hashpoint import HashPoint
from .hashresolver import HashResolver
from .mentionable import Mentionable
from .resolvermetaorigin import ResolverMetaOrigin

__all__ = ('ExtendableResolver',)

Mentioned = TypeVar('Mentioned')


class ExtendableResolver(HashResolver, abc.ABC):
    async def extend(self, hash_point: HashPoint[Mentioned]) -> 'ExtendableResolver':
        raise NotImplementedError

    async def migrate(self, hash_point: HashPoint[Mentioned]) -> HashPoint[Mentioned]:
        assert isinstance(hash_point, HashPoint)
        return ResolverMetaOrigin(await self.extend(hash_point)).hash_point(hash_point.factory, hash_point.point)

    async def migrate_resolved(self, mentioned: Mentioned) -> Mentioned:
        assert isinstance(mentioned, Mentionable)
        return await (await self.migrate(HashPoint.of(mentioned))).resolve()
