import abc
from typing import Generic, Type, TypeVar

from .hashresolver import HashResolver
from .mentionable import Mentionable
from .rainbow_factory import RainbowFactory

__all__ = ('StaticMentionable', 'StaticFactory',)

StaticMentioned = TypeVar('StaticMentioned')


class StaticMentionable(Mentionable, abc.ABC):
    @classmethod
    def from_bytes(cls: Type[StaticMentioned], source: bytes, resolver: HashResolver) -> StaticMentioned:
        raise NotImplementedError

    def __factory__(self) -> RainbowFactory[StaticMentioned]:
        return self.factory()

    @classmethod
    def factory(cls: Type[StaticMentioned]) -> RainbowFactory[StaticMentioned]:
        return StaticFactory(cls)


class StaticFactory(RainbowFactory[StaticMentioned], Generic[StaticMentioned]):
    def __init__(self, cls: Type[StaticMentioned]):
        assert issubclass(cls, StaticMentionable)
        self.cls: Type[StaticMentioned] = cls

    def from_bytes(self, source: bytes, resolver: HashResolver) -> StaticMentioned:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        if issubclass(self.cls, StaticMentionable):
            return self.cls.from_bytes(source, resolver)
        else:
            raise TypeError
