from typing import Callable

from .instrumentation import Instrumentation

__all__ = ('EntryExit',)


class EntryExit(Instrumentation):
    def __init__(self, target, methodname: str, point: Callable[[], tuple[float, int]]):
        assert callable(point)
        assert isinstance(methodname, str)
        super().__init__(target, methodname)
        self.point = point
        self.entry_log: list[tuple[float, int]] = []
        self.exit_log: list[tuple[float, int]] = []

    async def instrument(self, method, *args, **kwargs):
        self.entry_log.append(self.point())
        result = await method(*args, **kwargs)
        self.exit_log.append(self.point())
        return result
