from typing import TypeVar

from rainbowadn.core import *

__all__ = ('CachingResolver',)

Mentioned = TypeVar('Mentioned')


class CachingResolver(ExtendableResolver):
    def __init__(self, resolver: HashResolver):
        assert isinstance(resolver, HashResolver)
        self.resolver = resolver
        self.cache: dict[bytes, tuple[bytes, CachingResolver]] = {}

    async def resolve(self, point: bytes) -> tuple[bytes, HashResolver]:
        assert isinstance(point, bytes)
        if point not in self.cache:
            resolved, resolver = await self.resolver.resolve(point)
            assert isinstance(resolved, bytes)
            assert isinstance(resolver, HashResolver)
            self.cache[point] = resolved, CachingResolver(resolver)
        return self.cache[point]

    async def extend(self, hash_point: HashPoint[Mentioned]) -> ExtendableResolver:
        assert isinstance(hash_point, HashPoint)
        if isinstance(self.resolver, ExtendableResolver):
            return CachingResolver(await self.resolver.extend(hash_point))
        else:
            raise TypeError
