import asyncio
from typing import Callable, TypeVar

from rainbowadn.core import *

__all__ = ('DelayedResolver',)

Mentioned = TypeVar('Mentioned')


class DelayedResolver(ExtendableResolver):
    def __init__(self, resolver: HashResolver, delay: Callable[[], float]):
        assert isinstance(resolver, HashResolver)
        assert callable(delay)
        self.resolver = resolver
        self.delay = delay

    async def sleep(self):
        await asyncio.sleep(self.delay())

    async def resolve(self, point: bytes) -> tuple[bytes, 'HashResolver']:
        assert isinstance(point, bytes)
        await self.sleep()
        resolved, resolver = await self.resolver.resolve(point)
        assert isinstance(resolved, bytes)
        assert isinstance(resolver, HashResolver)
        return resolved, DelayedResolver(resolver, self.delay)

    async def extend(self, hash_point: HashPoint[Mentioned]) -> ExtendableResolver:
        assert isinstance(hash_point, HashPoint)
        if isinstance(self.resolver, ExtendableResolver):
            return DelayedResolver(await self.resolver.extend(hash_point), self.delay)
        else:
            raise TypeError
