from typing import Generic, TypeVar

from rainbowadn.chain import *
from rainbowadn.chain.reduction import *
from rainbowadn.collection.linear import *
from rainbowadn.core import *
from rainbowadn.nullability import *
from .bankprotocol import BankProtocol
from .bankstate import BankState
from .transaction import *

__all__ = ('BankChain',)

BlockType = TypeVar('BlockType')


class BankChain(Generic[BlockType]):
    def __init__(
            self,
            chain: ChainCollectionInterface[
                BlockType,
                NullableReference[Stack[Transaction]],
                BankState,
            ]
    ):
        assert isinstance(chain, ChainCollectionInterface)
        self.chain = chain
        self.reference = self.chain.reference
        assert isinstance(self.reference, NullableReference)

    @classmethod
    def empty(
            cls,
            factory: AbstractReductionChainMetaFactory[BlockType, NullableReference[Stack[Transaction]], BankState]
    ) -> 'BankChain[BlockType]':
        assert isinstance(factory, AbstractReductionChainMetaFactory)
        return cls(
            factory.factory(
                NullableReferenceFactory(StackFactory(Transaction.factory()).loose()).loose(),
                BankState.factory(),
                BankProtocol(),
            ).empty()
        )

    @classmethod
    def from_reference(
            cls,
            factory: AbstractReductionChainMetaFactory[BlockType, NullableReference[Stack[Transaction]], BankState],
            reference: NullableReference[
                BlockType
            ]
    ) -> 'BankChain[BlockType]':
        assert isinstance(factory, AbstractReductionChainMetaFactory)
        assert isinstance(reference, NullableReference)
        return cls(
            factory.factory(
                NullableReferenceFactory(StackFactory(Transaction.factory()).loose()).loose(),
                BankState.factory(),
                BankProtocol(),
            ).from_reference(
                reference
            )
        )

    async def add(self, transactions: NullableReference[Stack[Transaction]]) -> 'BankChain[BlockType]':
        assert isinstance(transactions, NullableReference)
        return BankChain(
            await self.chain.add(HashPoint.of(transactions))
        )

    async def adds(self, transactions: list[Transaction]) -> 'BankChain[BlockType]':
        assert isinstance(transactions, list)
        return await self.add(
            Stack.off(
                Transaction.factory(),
                reversed(transactions)
            )
        )

    async def verify(self) -> bool:
        return await self.chain.verify()
