from rainbowadn.atomic import *
from rainbowadn.chain.reduction import *
from rainbowadn.collection.keymetadata import *
from rainbowadn.collection.linear import *
from rainbowadn.collection.trees.binary import *
from rainbowadn.core import *
from rainbowadn.nullability import *
from .bankstate import BankState
from .subject import Subject
from .transaction import *

__all__ = ('BankProtocol',)


class BankProtocol(ReductionProtocol[NullableReference[Stack[Transaction]], BankState]):
    async def reduce(
            self,
            reduce: Reducible[NullableReference[Stack[Transaction]], BankState]
    ) -> ReductionResult[NullableReference[Stack[Transaction]], BankState]:
        assert isinstance(reduce, Reducible)
        bank_state: BankState
        reference: NullableReference[Stack[Transaction]]
        bank_state, reference = await gather(
            reduce.accumulator.resolve(),
            reduce.reductor.resolve(),
        )
        assert isinstance(bank_state, BankState)
        assert isinstance(reference, NullableReference)
        if reference.null():
            return Reduced(HashPoint.of(bank_state.without_miner()))
        else:
            transactions: Stack[Transaction] = await reference.resolve()
            assert isinstance(transactions, Stack)
            return Reducible(
                HashPoint.of(transactions.previous),
                HashPoint.of(await bank_state.push(transactions.element))
            )

    def initial(self, factory: RainbowFactory[BankState]) -> HashPoint[BankState]:
        assert isinstance(factory, RainbowFactory)
        return HashPoint.of(
            BankState(
                NullableReference(
                    Null(),
                    BinaryTreeFactory(KeyMetadataFactory(StaticFactory(Coin), StaticFactory(Integer)).loose()).loose()
                ),
                NullableReference(
                    Null(),
                    BinaryTreeFactory(KeyMetadataFactory(StaticFactory(Coin), StaticFactory(Integer)).loose()).loose()
                ),
                NullableReference(Null(), Subject.factory()),
                HashPoint.of(Integer(0))
            )
        )

    async def header_filter(self, state: HashPoint[BankState]) -> HashPoint[BankState]:
        assert isinstance(state, HashPoint)
        return HashPoint.of(await (await state.resolve()).without_miner().advance())
