import nacl.bindings
import nacl.exceptions
import nacl.signing

from rainbowadn.atomic import *
from rainbowadn.core import *
from .subject import Subject

__all__ = ('BadSignature', 'Signature',)


class BadSignature(nacl.exceptions.BadSignatureError):
    pass


class Signature(Atomic):
    def __init__(self, source: bytes):
        assert isinstance(source, bytes)
        assert_eq(len(source), nacl.bindings.crypto_sign_BYTES)
        self.source = source

    @classmethod
    def sign(cls, key: nacl.signing.SigningKey, hash_point: HashPoint) -> 'Signature':
        assert isinstance(key, nacl.signing.SigningKey)
        assert isinstance(hash_point, HashPoint)
        return cls(
            key.sign(hash_point.point).signature
        )

    @classmethod
    def _from_bytes(cls, source: bytes) -> 'Signature':
        assert isinstance(source, bytes)
        return cls(source)

    def __bytes__(self):
        return self.source

    def verify(self, subject: Subject, hash_point: HashPoint) -> bool:
        assert isinstance(subject, Subject)
        assert isinstance(hash_point, HashPoint)
        try:
            subject.verify_key.verify(hash_point.point, self.source)
            return True
        except nacl.exceptions.BadSignatureError:
            raise BadSignature

    def __str__(self):
        return f'(signature)'
