from typing import Generic, TypeVar

from rainbowadn.chain import *
from rainbowadn.core import *
from .wrisbtindex import *
from .wrisbtparametres import WrisbtParametres
from .wrisbtprotocol import WrisbtProtocol
from .wrisbtroot import *

__all__ = ('WrisbtChainProtocol',)

TargetType = TypeVar('TargetType')


class WrisbtChainProtocol(
    BlockChainProtocol[
        TargetType,
        WrisbtIndex,
        WrisbtRoot
    ],
    Generic[TargetType]
):
    def __init__(
            self,
            total_factory: RainbowFactory[TargetType],
            keymin: int
    ):
        assert isinstance(total_factory, RainbowFactory)
        assert isinstance(keymin, int)
        assert keymin > 0
        self.keymin = keymin
        super().__init__(
            WrisbtProtocol(keymin),
            total_factory,
            WrisbtIndexFactory(keymin)
        )
        self.total_factory = total_factory

    def actual_state_factory(self) -> RainbowFactory[WrisbtRoot]:
        return WrisbtRootFactory(WrisbtParametres(self.keymin, HashPoint.HASH_LENGTH))

    async def actual_state(self, state: WrisbtIndex) -> HashPoint[WrisbtRoot]:
        assert isinstance(state, WrisbtIndex)
        return state.total
