from typing import Iterable

from rainbowadn.core import *
from .wrisbtparametres import WrisbtParametres
from .wrisbtroot import *

__all__ = ('WrisbtIndex', 'WrisbtIndexFactory',)


class WrisbtIndex(RecursiveMentionable):
    def __init__(
            self,
            total: HashPoint[WrisbtRoot],
            delta: HashPoint[WrisbtRoot],
            keymin: int
    ):
        assert isinstance(total, HashPoint)
        assert isinstance(delta, HashPoint)
        assert isinstance(keymin, int)
        self.total = total
        self.delta = delta
        assert keymin > 0
        self.keymin = keymin

    def points(self) -> Iterable[HashPoint]:
        return [self.total, self.delta]

    def __bytes__(self):
        return bytes(self.total) + bytes(self.delta)

    def __factory__(self) -> RainbowFactory['WrisbtIndex']:
        return WrisbtIndexFactory(self.keymin)

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        total_str, delta_str = await gather(
            hash_point_format(self.total, tab),
            hash_point_format(self.delta, tab),
        )
        return f'(index)' \
               f'{tabulate(tab)}{total_str}' \
               f'{tabulate(tab)}{delta_str}'


class WrisbtIndexFactory(RainbowFactory[WrisbtIndex]):
    def __init__(self, keymin: int):
        assert isinstance(keymin, int)
        assert keymin > 0
        self.keymin = keymin
        self.root_factory: RainbowFactory[WrisbtRoot] = WrisbtRootFactory(
            WrisbtParametres(keymin, HashPoint.HASH_LENGTH)
        )
        assert isinstance(self.root_factory, RainbowFactory)

    def from_bytes(self, source: bytes, resolver: HashResolver) -> WrisbtIndex:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return WrisbtIndex(
            ResolverOrigin(self.root_factory, source[:HashPoint.HASH_LENGTH], resolver).hash_point(),
            ResolverOrigin(self.root_factory, source[HashPoint.HASH_LENGTH:], resolver).hash_point(),
            self.keymin
        )
